#!/usr/bin/php
<?php
require_once('ShellTools.class');

class RepairLeadingSpaces {
    /**
     * True = tab goes to the next $tabWidth character column.
     *     e.g. three spaces then a tab (8) will still go to column 8
     * False = tab always inserts $tabWidth characters.
     *     e.g. three spaces then a tab (8) will go to column 11
     */
    var $tabToColumn = true;
    var $tabWidth = 8;
    var $fileExtensions = array('php', 'inc', 'class');
    var $backupExtension = '.bak';

    var $saveChanges = false;
    var $spacePattern = '/^([\t]* +[\t]+|[\t]* {8,})/';
    var $files = array();
    var $_addedSomething = false;

    var $shell;

    function RepairLeadingSpaces() {
	$this->shell = new ShellTools;

	$commands = array(
	    ''          => array(&$this, 'argumentAdd'), 
	    's'         => array(&$this, 'argumentSave'), 
	    'save'      => array(&$this, 'argumentSave'), 
	    'h'         => array(&$this, 'argumentHelp'), 
	    'help'      => array(&$this, 'argumentHelp'), 
	    'b'         => array(&$this, 'argumentBackup'), 
	    'backup'    => array(&$this, 'argumentBackup'), 
	    'n'         => array(&$this, 'argumentNoBackup'), 
	    'nobackup'  => array(&$this, 'argumentNoBackup'), 
	    't'         => array(&$this, 'argumentTabWidth'), 
	    'tabwidth'  => array(&$this, 'argumentTabWidth'), 
	    'e'         => array(&$this, 'argumentTabColumn'), 
	    'tabcolumn' => array(&$this, 'argumentTabColumn'), 
	    'f'         => array(&$this, 'argumentFileExtension'), 
	    'fileextension' => array(&$this, 'argumentFileExtension'), 
	);

	$this->shell->getArguments($commands);

	if (!$this->_addedSomething && empty($this->files)) {
	    $this->addPath('.');
	}

	$exitStatus = 0;

	foreach ($this->files as $file) {
	    $contents = array_merge(array(''), file($file));
	    $changes = $this->repairLeadSpace($file, $contents);

	    if (empty($changes)) {
		continue;
	    } else if ($this->saveChanges) {
		if (!empty($this->backupExtension)) {
		    $backupExtension = $this->backupExtension;
		    $backupStatus = `cp "$file" "$file$backupExtension"`;
		    if (!empty($backupStatus)) {
			continue;
		    }
		}

		$fileHandle = fopen($file, 'w');
		if (!$fileHandle) {
		    $exitStatus = 1;
		    $this->shell->error('Could not open file ' . $file);
		} else {
		    $fileWrite = fwrite($fileHandle, implode('', $contents));
		    if (!$fileWrite) {
			$exitStatus = 1;
			$this->shell->error('Could not save file ' . $file);
		    }
		    fclose($fileHandle);
		}
	    } else {
		echo $file . ': ' . implode(', ', $changes) . "\n";
	    }
	}

	exit($exitStatus);
    }

    function argumentHelp() {
	global $argv;

	echo 'Usage: ' . $argv[0] . ' [options] [[<path>|<file>]...]' . "\n" . 
	    '  -h              This help information.                            ' . "\n" . 
	    '  -s              Repair errors and save.                           ' . "\n" . 
	    '  -b .ext         Make a backup using "ext". Default is ".bak".     ' . "\n" . 
	    '  -n              Don\'t make a backup when saving.                 ' . "\n" . 
	    '                  While it wouldn\'t make sense to use this with -b,' . "\n" . 
	    '                  it is allowed. Using this option after -b will    ' . "\n" . 
	    '                  override backups, and vice versa.                 ' . "\n" . 
	    '  -t #            Set tab width. Defaults to 8.                     ' . "\n" . 
	    '  -e true|false   Set if tabs should be interpretted as expansions  ' . "\n" . 
	    '                  to the next column. Default is true.              ' . "\n" . 
	    '  -f string       Comma-delimited string of extensions to search.   ' . "\n" . 
	    '                  Defaults to "php,inc,class"                       ' . "\n" . 
	    '                                                                    ' . "\n" . 
	    'Examples:                                                           ' . "\n" . 
	    "\t" . 'Set tab width to 4 and search .php, .inc, .class, and .tpl files:' . "\n" . 
	    "\t\t" . $argv[0] . ' -t 4 -f php,inc,class,tpl' . "\n" . 
	    "\t" . 'Save changes and make a backup as *.backup:                      ' . "\n" . 
	    "\t\t" . $argv[0] . ' -s -b .backup' . "\n" . 
	    "\t" . 'Save, backup, tab width, and files in one.                       ' . "\n" . 
	    "\t" . 'Note the parameters follow the order of the arguments:           ' . "\n" . 
	    "\t\t" . $argv[0] . ' -sbtf .backup 4 tpl' . "\n";
	exit(0);
    }

    function argumentSave() {
	$this->saveChanges = true;
    }

    function argumentBackup($argn) {
	global $argv;

	if (empty($argv[$argn + 1])) {
	    $this->shell->error('Error: no backup extension specified');
	} else {
	    $this->backupExtension = $argv[$argn + 1];
	    return 1;
	}
    }

    function argumentNoBackup($argn) {
	$this->backupExtension = '';
    }

    function argumentTabWidth($argn) {
	global $argv;
	$argn++;

	if (empty($argv[$argn]) || !is_numeric($argv[$argn])) {
	    $this->shell->error('Error: no tab width specified');
	} else {
	    $tabWidth = (int)$argv[$argn];
	    $this->spacePattern = str_replace($this->tabWidth, $tabWidth, $this->spacePattern);
	    $this->tabWidth = $tabWidth;
	    return 1;
	}
    }

    function argumentTabColumn($argn) {
	global $argv;

	if (empty($argv[$argn + 1]) || !in_array($argv[$argn + 1], array('true', 'false'))) {
	    $this->shell->error('Error: no tab column boolean specified.');
	} else {
	    $this->tabToColumn = ($argv[$argn + 1] == 'true') ? true : false;
	    return 1;
	}
    }

    function argumentFileExtension($argn) {
	global $argv;

	if (empty($argv[$argn + 1])) {
	    $this->shell->error('Error: No file extension(s) specified.');
	} else {
	    $this->fileExtensions = array_filter(explode(',', $argv[$argn + 1]));
	    return 1;
	}
    }

    function argumentAdd($argn) {
	global $argv;

	$this->_addedSomething = true;
	if (is_dir($argv[$argn])) {
	    $this->addPath($argv[$argn]);
	} else {
	    $this->addFile($argv[$argn]);
	}
    }

    function addPath($path) {
	if (empty($this->findMatchFiles)) {
	    $this->findMatchFiles = '-name "*.' . 
		implode('" -o -name "*.', $this->fileExtensions) . '"';
	}
	$findMatchFiles = $this->findMatchFiles;

	$find = `find "$path" -type f -name ".svn" -prune -o $findMatchFiles`;
	$this->files = array_merge($this->files, array_filter(explode("\n", $find)));
    }

    function addFile($file) {
	$this->files = array_merge($this->files, array($file));
    }

    function repairLeadSpace($file, &$contents) {
	$preg_result = preg_grep($this->spacePattern, $contents);
	if (empty($preg_result)) {
	    return false;
	}

	if (!$this->saveChanges) {
	    return array_keys($preg_result);
	}

	foreach ($preg_result as $number => $line) {
	    $strlen = strlen($line);
	    $spaces = 0;
	    $blankLine = $returnLine = false;

	    for ($i = 0; $i < $strlen; $i++) {
		switch ($line{$i}) {
		case ' ':
		    $spaces++;
		    break;
		case "\t":
		    if ($this->tabToColumn) {
			$tabPosition = $spaces % $this->tabWidth;
			$spaces += $this->tabWidth - $tabPosition;
		    } else {
			$spaces += $this->tabWidth;
		    }
		    break;
		case "\r":
		    $returnLine = true;
		case "\n":
		    $blankLine = true;
		    break 2;
		default:
		    break 2;
		}
	    }

	    if ($blankLine) {
		$newLine = ($returnLine) ? "\r\n" : "\n";
	    } else {
		$newLine = str_repeat("\t", floor($spaces / $this->tabWidth));
		$newLine .= str_repeat(' ', $spaces % $this->tabWidth);
		$newLine .= ltrim($line);
	    }

	    $contents[$number] = $newLine;
	}

	return true;
    }
}

$repairLeadingSpaces = new RepairLeadingSpaces;

?>